! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! June, 2016

! Elastic code that generates data with a nonlinear solver

module elastic_mod

  use model_mod
  use clib_rtm_ispc
  use clib_lint_ispc
  use sep
  use source_mod
  use seis_mod_rtm
  use snap_mod_rtm
  use omp_lib
  use abc_mod

  implicit none
  integer, private :: nt, ntseis, itseis, nx, nz, xsource, zsource, abc, nf, nm, nrec, dtratio
  real, private    :: dt, dt2, dtseis, dx, dz, fp, aux, aux2, fc
  logical, private :: lame, snap, write_bkg
  integer, private :: zrec, xrecbeg, xrecend, xreclen
  integer, private, parameter :: T=1,F=0

  !Model
  real, pointer, dimension(:,:),private        :: m1,m2,m3
  real, allocatable, dimension(:,:), private   :: m1_x,m1_z
  real, allocatable, dimension(:,:), private   :: aux_m1_x,aux_m1_z,aux_m2,aux_m3,aux_m2m3
  !Data
  real, pointer, dimension(:,:),private :: Vx,Vz,sigmaxx,sigmazz,sigmaxz
  real, pointer, dimension(:,:),private :: Vx_data,Vz_data,sigmaxx_data,sigmazz_data,sigmaxz_data
  !Source
  real, allocatable, dimension(:), private   :: source
  
  !Temporary wavefields
  real, target, allocatable, dimension(:,:)   :: Vx1,Vz1,sigmaxx1,sigmazz1,sigmaxz1
  real, target, allocatable, dimension(:,:)   :: Vx2,Vz2,sigmaxx2,sigmazz2,sigmaxz2

  real, pointer, dimension(:,:), private      :: pVx_tmp,pVz_tmp,psigmaxx_tmp,psigmazz_tmp,psigmaxz_tmp
  real, pointer, dimension(:,:), private      :: pVx_cur,pVz_cur,psigmaxx_cur,psigmazz_cur,psigmaxz_cur
  real, pointer, dimension(:,:), private      :: pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new

  !Nodes thread split
  integer, private							  :: nodes
  integer, pointer, dimension(:), private     :: b,e

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elastic_init(nz_in,dz_in,nx_in,dx_in,dt_in,dtseis_in,dtratio_in,xsource_in,zsource_in,zrec_in,xrec_in,rec_length_in,fp_in,nt_in,ntseis_in,abc_in,m1_in,m2_in,m3_in,b_in,e_in,lame_in,snap_in,write_bkg_in)

    integer             :: nz_in, nx_in, dtratio_in, xsource_in, zsource_in, nt_in, ntseis_in, abc_in,zrec_in, xrec_in, rec_length_in
    real                :: dz_in, dx_in, dt_in, fp_in, dtseis_in
    logical             :: lame_in, snap_in, write_bkg_in
    !Background properties
    real, dimension(:,:),target  :: m1_in,m2_in,m3_in
    !Threads division
    integer, dimension(:), target :: b_in, e_in

    nz = nz_in
    dz = dz_in
    dt = dt_in
    dtseis = dtseis_in
    dtratio = dtratio_in
    nx = nx_in
    dx = dx_in
    xsource = xsource_in
    zsource = zsource_in
    fp = fp_in
    nt = nt_in
    ntseis = ntseis_in
    abc = abc_in
    zrec = zrec_in
    xrecbeg = xrec_in
    xreclen = rec_length_in
    lame = lame_in
    snap = snap_in
    b => b_in
    e => e_in
    write_bkg = write_bkg_in

    xrecend = xrecbeg + xreclen - 1

    nodes = size(b)

    !Input background model
    m1 => m1_in
    m2 => m2_in
    m3 => m3_in

    !Allocate model parameters
    allocate(m1_x(nz,nx))
    allocate(m1_z(nz,nx))
    allocate(aux_m1_x(nz,nx))
    allocate(aux_m1_z(nz,nx))
    allocate(aux_m2(nz,nx))
    allocate(aux_m3(nz,nx))
    allocate(aux_m2m3(nz,nx))

    !Allocate source
    allocate(source(nt))

    !Allocate Velocities and stresses
    !Temporary wavefields
    allocate(Vx1(nz,nx))
    allocate(Vz1(nz,nx))
    allocate(sigmaxx1(nz,nx))
    allocate(sigmazz1(nz,nx))
    allocate(sigmaxz1(nz,nx))
    allocate(Vx2(nz,nx))
    allocate(Vz2(nz,nx))
    allocate(sigmaxx2(nz,nx))
    allocate(sigmazz2(nz,nx))
    allocate(sigmaxz2(nz,nx))

    ! Write outputs to the terminal
    write(0,*)"============================================================="
    write(0,*)"MEMORY ALLOCATION FOR TEMPORARY ARRAYS"
    write(0,*)"============================================================="
    write(0,*)"INTERPOLATED DENSITY"
    write(0,*)"rho_x",size(m1_x)*4/1024,"KB"
    write(0,*)"rho_z",size(m1_z)*4/1024,"KB"
    write(0,*)"AUXILIARY PROPERTY ARRAYS"
    write(0,*)"aux_m1_x",size(aux_m1_x)*4/1024,"KB"
    write(0,*)"aux_m1_z",size(aux_m1_z)*4/1024,"KB"
    write(0,*)"aux_m2",size(aux_m2)*4/1024,"KB"
    write(0,*)"aux_m3",size(aux_m3)*4/1024,"KB"
    write(0,*)"aux_m2m3",size(aux_m2m3)*4/1024,"KB"
    write(0,*)"TEMPORARY ARRAYS"
    write(0,*)"Vx1",size(Vx1)*4/1024,"KB"
    write(0,*)"Vz1",size(Vz1)*4/1024,"KB"
    write(0,*)"sigmaxx1",size(sigmaxx1)*4/1024,"KB"
    write(0,*)"sigmazz1",size(sigmazz1)*4/1024,"KB"
    write(0,*)"sigmaxz1",size(sigmaxz1)*4/1024,"KB"
    write(0,*)"Vx2",size(Vx2)*4/1024,"KB"
    write(0,*)"Vz2",size(Vz2)*4/1024,"KB"
    write(0,*)"sigmaxx2",size(sigmaxx2)*4/1024,"KB"
    write(0,*)"sigmazz2",size(sigmazz2)*4/1024,"KB"
    write(0,*)"sigmaxz2",size(sigmaxz2)*4/1024,"KB"
    write(0,*)"============================================================="

    !Initialize values
    !Background properties
    m1_x  = 0.
    m1_z  = 0.

    !Source
    source=0.

    !Initialize subroutines
    !Model
    call model_init(nz,nx,dx)
    call model_op(m1,m2,m3,m1_x,m1_z)
    write(0,*)"MODEL_OP => OK"
    !Absorbing boundary
    call abc_init(nz,nx,dx,abc)

    aux = dt/dx
    aux2 = sqrt(float(ntseis)/float(nt))
    dt2 = dt/2.0 !The do loop actually steps at half time steps,
                 !so this is needed to sample the seismogram correctly

    !Snapshot
    if(snap) call snap_init(nz,dz,nx,dx,ntseis,dtseis)
    !Source
    call source_init(fp,dt2,dx,nz,nx,nt) !The source is also sampled
                                         !at half time steps

    call source_op2(source)

    aux_m1_x = aux*m1_x
    aux_m1_z = aux*m1_z
    if (lame) then
      aux_m2 = aux*1.0E6*m2
      aux_m3 = aux*1.0E6*m3
      aux_m2m3 = aux*(1.0E6*m2+2.0*1.0E6*m3)
    else
      aux_m2 = aux*m1*(m2*m2-2.0*m3*m3)
      aux_m3 = aux*m1*m3*m3
      aux_m2m3 = aux*m1*m2*m2
    endif

  end subroutine

  ! Function that is called by Main
  function elastic_op(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in) result(stat)
    real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in
    integer                         :: stat

    stat=1
    call elastic_op2(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in)
    stat=0
  end function

  ! Subroutine used by function
  subroutine elastic_op2(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in)
    real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in

    !General variables
    integer                         :: it, ix, iz

    b(1)=5
    e(nodes)=nx-5

    !Data
    Vx => Vx_in
    Vz => Vz_in
    sigmaxx => sigmaxx_in
    sigmazz => sigmazz_in
    sigmaxz => sigmaxz_in

!###################################################
! Elastic modeling of data d0
!###################################################
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"BEGIN OF TIME MODELING FOR WAVEFIELD"
    write(0,*)"============================================================="
    write(0,*)"Processing:"
    write(0,*)"0%           25%            50%           75%            100%"
    write(0,'(a)',advance="no")"["

    !Zero arrays
    Vx1      = 0.
    Vz1      = 0.
    sigmaxx1 = 0.
    sigmazz1 = 0.
    sigmaxz1 = 0.
    Vx2      = 0.
    Vz2      = 0.
    sigmaxx2 = 0.
    sigmazz2 = 0.
    sigmaxz2 = 0.

    !Initialize pointers
    pVx_new => Vx1
    pVz_new => Vz1
    psigmaxx_new => sigmaxx1
    psigmazz_new => sigmazz1
    psigmaxz_new => sigmaxz1
    pVx_cur => Vx2
    pVz_cur => Vz2
    psigmaxx_cur => sigmaxx2
    psigmazz_cur => sigmazz2
    psigmaxz_cur => sigmaxz2
    
! First time steps
    !it=1
    !Boundary condition, all terms are zero

    !it=2
    it=2
    itseis  = it*(dt2/dtseis)

    call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
    call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
    
    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)

    if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

    !Update pointers
	pVx_tmp => pVx_cur
	pVz_tmp => pVz_cur
	psigmaxx_tmp => psigmaxx_cur
	psigmazz_tmp => psigmazz_cur
	psigmaxz_tmp => psigmaxz_cur
	pVx_cur => pVx_new  
	pVz_cur => pVz_new 
	psigmaxx_cur => psigmaxx_new
	psigmazz_cur => psigmazz_new
	psigmaxz_cur => psigmaxz_new
	pVx_new => pVx_tmp
	pVz_new => pVz_tmp
	psigmaxx_new => psigmaxx_tmp
	psigmazz_new => psigmazz_tmp
	psigmaxz_new => psigmaxz_tmp

    !it=3
    it=3
    itseis  = it*(dt2/dtseis)
    
    !Vx  
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(F,F,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO

    !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(F,F,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
       call Zero_array_c(nx,nz,b(ix),e(ix),psigmaxx_new)
       call Zero_array_c(nx,nz,b(ix),e(ix),psigmazz_new)
    enddo
!$OMP END PARALLEL DO 
    !sources
    call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
    call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3)
    !sigmazz
      call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2)
    !sigmaxz
      call Bz_c(F,F,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2)
    !sigmazz
      call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3)
    !sigmaxz
      call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)

    if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

    !Update pointers
	pVx_tmp => pVx_cur
	pVz_tmp => pVz_cur
	psigmaxx_tmp => psigmaxx_cur
	psigmazz_tmp => psigmazz_cur
	psigmaxz_tmp => psigmaxz_cur
	pVx_cur => pVx_new  
	pVz_cur => pVz_new 
	psigmaxx_cur => psigmaxx_new
	psigmazz_cur => psigmazz_new
	psigmaxz_cur => psigmaxz_new
	pVx_new => pVx_tmp
	pVz_new => pVz_tmp
	psigmaxx_new => psigmaxx_tmp
	psigmazz_new => psigmazz_tmp
	psigmaxz_new => psigmaxz_tmp

    !it=4 until it=nt
    do it=4,nt-1
      itseis  = it*(dt2/dtseis)
      
      !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
      enddo
!$OMP END PARALLEL DO
      !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
      enddo
!$OMP END PARALLEL DO
      call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
      call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
      !sigmaxx
        call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3) !D
      !sigmazz
        call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2) !F
      !sigmaxz
        call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3) !H
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
      !sigmaxx
        call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2) !E
      !sigmazz
        call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3) !G
      !sigmaxz
        call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3) !J
      enddo
!$OMP END PARALLEL DO
      
      call K(pVx_new)
      call K(pVz_new)
      call K(psigmaxx_new)
      call K(psigmazz_new)
      call K(psigmaxz_new)

      if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

      !Update pointers
	  pVx_tmp => pVx_cur
	  pVz_tmp => pVz_cur
	  psigmaxx_tmp => psigmaxx_cur
	  psigmazz_tmp => psigmazz_cur
	  psigmaxz_tmp => psigmaxz_cur
	  pVx_cur => pVx_new  
	  pVz_cur => pVz_new 
	  psigmaxx_cur => psigmaxx_new
	  psigmazz_cur => psigmazz_new
	  psigmaxz_cur => psigmaxz_new
	  pVx_new => pVx_tmp
	  pVz_new => pVz_tmp
	  psigmaxx_new => psigmaxx_tmp
	  psigmazz_new => psigmazz_tmp
	  psigmaxz_new => psigmaxz_tmp

      !Write to terminal
      if (mod(it,(nt/59))==0) then
        write(0,'(a)',advance="no")"#"
        close(0)
      endif

    enddo
    write(0,'(a)',advance="yes")"]"

    if(write_bkg) call seismo2(nz,dz,nx,dx,dt,pvx_new,pvx_cur,pvz_new,pvz_cur,psigmaxx_new,psigmaxx_cur,psigmazz_new,psigmazz_cur,psigmaxz_new,psigmaxz_cur)

    !Deallocate arrays
    deallocate(m1_x,m1_z)
    deallocate(source)
    deallocate(Vx1,Vz1,sigmaxx1,sigmazz1,sigmaxz1,Vx2,Vz2,sigmaxx2,sigmazz2,sigmaxz2)

  end subroutine

!#############################################################
! Elastic modeling of data from another data (debugging tool)
!#############################################################

  ! Function that is called by Main
  function elastic_op3(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in,Vx_out,Vz_out,sigmaxx_out,sigmazz_out,sigmaxz_out) result(stat)
    real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in
    real,dimension(:,:),target      :: Vx_out,Vz_out,sigmaxx_out,sigmazz_out,sigmaxz_out
    integer                         :: stat

    stat=1
    call elastic_op4(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in,Vx_out,Vz_out,sigmaxx_out,sigmazz_out,sigmaxz_out)
    stat=0
  end function

  ! Subroutine used by function
  subroutine elastic_op4(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in,Vx_out,Vz_out,sigmaxx_out,sigmazz_out,sigmaxz_out)
    real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in
    real,dimension(:,:),target      :: Vx_out,Vz_out,sigmaxx_out,sigmazz_out,sigmaxz_out
    
    !General variables
    integer                         :: it, ix, iz

    !Data
    Vx_data => Vx_in
    Vz_data => Vz_in
    sigmaxx_data => sigmaxx_in
    sigmazz_data => sigmazz_in
    sigmaxz_data => sigmaxz_in   
    Vx => Vx_out
    Vz => Vz_out
    sigmaxx => sigmaxx_out
    sigmazz => sigmazz_out
    sigmaxz => sigmaxz_out

!###################################################
! Elastic modeling of data d0
!###################################################
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"BEGIN OF TIME MODELING FOR WAVEFIELD"
    write(0,*)"============================================================="
    write(0,*)"Processing:"
    write(0,*)"0%           25%            50%           75%            100%"
    write(0,'(a)',advance="no")"["

    !Zero arrays
    Vx1      = 0.
    Vz1      = 0.
    sigmaxx1 = 0.
    sigmazz1 = 0.
    sigmaxz1 = 0.
    Vx2      = 0.
    Vz2      = 0.
    sigmaxx2 = 0.
    sigmazz2 = 0.
    sigmaxz2 = 0.

    !Initialize pointers
    pVx_new => Vx1
    pVz_new => Vz1
    psigmaxx_new => sigmaxx1
    psigmazz_new => sigmazz1
    psigmaxz_new => sigmaxz1
    pVx_cur => Vx2
    pVz_cur => Vz2
    psigmaxx_cur => sigmaxx2
    psigmazz_cur => sigmazz2
    psigmaxz_cur => sigmaxz2
    
! First time steps
    !it=1
    !Boundary condition, all terms are zero
    it=1
    if(snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)
    
    !it=2
    it=2
    itseis  = it*(dt2/dtseis)

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(F,F,xreclen,dt2,dtseis,Vx_data(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,F,xreclen,dt2,dtseis,Vz_data(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmaxx_data(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmazz_data(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmaxz_data(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

   call K(pVx_new)
   call K(pVz_new)
   call K(psigmaxx_new)
   call K(psigmazz_new)
   call K(psigmaxz_new)

    if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS
 
    !Update pointers
	pVx_tmp => pVx_cur
	pVz_tmp => pVz_cur
	psigmaxx_tmp => psigmaxx_cur
	psigmazz_tmp => psigmazz_cur
	psigmaxz_tmp => psigmaxz_cur
	pVx_cur => pVx_new  
	pVz_cur => pVz_new 
	psigmaxx_cur => psigmaxx_new
	psigmazz_cur => psigmazz_new
	psigmaxz_cur => psigmaxz_new
	pVx_new => pVx_tmp
	pVz_new => pVz_tmp
	psigmaxx_new => psigmaxx_tmp
	psigmazz_new => psigmazz_tmp
	psigmaxz_new => psigmaxz_tmp
    
    !it=3
    it=3
    itseis  = it*(dt2/dtseis)
    
    !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(F,F,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
    !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(F,F,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Az_c(F,F,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3)
    !sigmazz
      call Az_c(F,F,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2)
    !sigmaxz
      call Bz_c(F,F,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2)
    !sigmazz
      call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3)
    !sigmaxz
      call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vx_data(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vz_data(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxx_data(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmazz_data(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxz_data(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)


    if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

    !Update pointers
	pVx_tmp => pVx_cur
	pVz_tmp => pVz_cur
	psigmaxx_tmp => psigmaxx_cur
	psigmazz_tmp => psigmazz_cur
	psigmaxz_tmp => psigmaxz_cur
	pVx_cur => pVx_new  
	pVz_cur => pVz_new 
	psigmaxx_cur => psigmaxx_new
	psigmazz_cur => psigmazz_new
	psigmaxz_cur => psigmaxz_new
	pVx_new => pVx_tmp
	pVz_new => pVz_tmp
	psigmaxx_new => psigmaxx_tmp
	psigmazz_new => psigmazz_tmp
	psigmaxz_new => psigmaxz_tmp

    !it=4 until it=nt
    do it=4,nt-1
      itseis  = it*(dt2/dtseis)
        
      !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
        call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
        call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
      enddo
!$OMP END PARALLEL DO
      !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
        call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
        call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
      enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      !sigmaxx
        call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3) !D
      !sigmazz
        call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2) !F
      !sigmaxz
        call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3) !H
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      !sigmaxx
        call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2) !E
      !sigmazz
        call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3) !G
      !sigmaxz
        call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3) !J
      enddo
!$OMP END PARALLEL DO

!$OMP SECTIONS
!$OMP SECTION
      call lint_serial_c(F,T,xreclen,dt2,dtseis,Vx_data(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(F,T,xreclen,dt2,dtseis,Vz_data(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxx_data(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmazz_data(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxz_data(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

      call K(pVx_new)
      call K(pVz_new)
      call K(psigmaxx_new)
      call K(psigmazz_new)
      call K(psigmaxz_new)
      
      if (mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

      ! if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

!$OMP SECTIONS
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
      call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

      !Update pointers
	  pVx_tmp => pVx_cur
	  pVz_tmp => pVz_cur
	  psigmaxx_tmp => psigmaxx_cur
	  psigmazz_tmp => psigmazz_cur
	  psigmaxz_tmp => psigmaxz_cur
	  pVx_cur => pVx_new  
	  pVz_cur => pVz_new 
	  psigmaxx_cur => psigmaxx_new
	  psigmazz_cur => psigmazz_new
	  psigmaxz_cur => psigmaxz_new
	  pVx_new => pVx_tmp
	  pVz_new => pVz_tmp
	  psigmaxx_new => psigmaxx_tmp
	  psigmazz_new => psigmazz_tmp
	  psigmaxz_new => psigmaxz_tmp

      !Write to terminal
      if (mod(it,(nt/59))==0) then
        write(0,'(a)',advance="no")"#"
        close(0)
      endif

    enddo

    !nt
	it=nt
	if(snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)
    
    write(0,'(a)',advance="yes")"]"

    !Deallocate arrays
    deallocate(m1_x,m1_z)
    deallocate(source)
    deallocate(Vx1,Vz1,sigmaxx1,sigmazz1,sigmaxz1,Vx2,Vz2,sigmaxx2,sigmazz2,sigmaxz2)

  end subroutine

end module
