!!$=head1 NAME
!!$Elastic modeling
!!$
!!$=head1 DESCRIPTION
!!$
!!$ 2D modeling program that creates elastic data from a set of 
!!$ velocities (vp,vs) and density models or Lame (lambda, mu)
!!$ and density models. Outputs data in the form of particle
!!$ velocities, normal and shear stresses.
!!$
!!$=head1 USAGE
!!$
!!$elastic.x  <input 1 >/dev/null [PARAMETERS]
!!$
!!$=head1 INPUT PARAMETERS
!!$
!!$=over 3
!!$
!!$=item lame
!!$
!!$0 for vp,vs,rho
!!$
!!$1 for lambda,mu,rho
!!$
!!$=item input 1
!!$
!!$if lame=0, .H file of Vp model in m/s
!!$
!!$if lame=1, .H file of lambda model in MPa
!!$
!!$=item mu
!!$
!!$.H file of Mu model in MPa (for lame=1)
!!$
!!$=item vs
!!$
!!$.H file of Vs model in m/s (for lame=0)
!!$
!!$=item rho
!!$
!!$.H file of density model in kg/m3
!!$
!!$=item Vx_rec
!!$
!!$.H file of velocity x data
!!$
!!$=item Vz_rec
!!$
!!$.H file of velocity z data
!!$
!!$=item sigmaxx_rec
!!$
!!$.H file of normal stress in x data
!!$
!!$=item sigmazz_rec
!!$
!!$.H file of normal stress in z data
!!$
!!$=item sigmaxz_rec
!!$
!!$.H file of shear stress data
!!$
!!$=item xsource
!!$
!!$position of the source in the x coordinate in grid points
!!$
!!$=item zsource
!!$
!!$position of the source in the z coordinate in grid points
!!$
!!$=item Ttime
!!$
!!$total propagation time in seconds
!!$
!!$=item fp
!!$
!!$peak frequency of the Ricker type source
!!$
!!$=item abc
!!$
!!$width of absorbing boundary condition in grid points
!!$
!!$=item zrec
!!$
!!$depth of receiver line in grid points
!!$
!!$=back
!!$
!!$=head1 OPTIONAL PARAMETERS
!!$
!!$=over 3
!!$
!!$=item dtseis
!!$
!!$sampling of output seismogram in seconds [default dtseis=0.002]
!!$
!!$=item snap
!!$
!!$if 1, saves all data wavefields at every dtseis sample [default snap=0]
!!$
!!$=item rec_source
!!$
!!$if 1, reads data as input for the source use vx_in, vz_in, sigmaxx_in, sigmazz_in
!!$and sigmaxz_in to choose the .H files for input [default rec_source=0]
!!$
!!$=item dtflag
!!$
!!$if 1, reads time.txt to choose the time sampling for propagation and ratio between
!!$time sampling and seismogram sampling, instead of calculating them from stability
!!$condition [default dtflag=0]  
!!$
!!$=back
!!$
!!$=head1 ABOUT
!!$Program created by Gustavo Catao Alves at
!!$the Stanford Exploration Project
!!$
!!$Updated in March, 3rd, 2016
!!$
!!$=cut

! Main program to execute elastic modeling
program elastic_program_rec

  use sep
  use elastic_mod
  use snap_mod_rtm
  use seis_mod_rtm
  use random_mod

  use omp_lib

  implicit none

  integer                           :: nz,nx,xsource,zsource,nt,abc,zrec,xrec,rec_length,ntseis,dtratio
  integer                           :: stat, ix, iz
  real                              :: dz,oz,dx,ox,dt,fp,Tt,ot,dtseis
  real, allocatable, dimension(:,:) :: rho,lambda,mu, vp, vs
  real, allocatable, dimension(:,:) :: Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec
  real, allocatable, dimension(:,:) :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in
  real                              :: lval,mval,rval,vtemp,vmax
  logical                           :: lame, dtflag, rec_source, snap, write_bkg

  !Variables to measure runtime
  real*8                            :: start_time,end_time

  !OMP settings
  integer                             :: node, nodes
  integer, allocatable, dimension(:)	:: b,e
  start_time = omp_get_wtime()

  call sep_init(SOURCE)

  ! Read SEP header parameter
  ! Model parameters
  call from_history("n1",nz)
  call from_history("o1",oz)
  call from_history("d1",dz)
  call from_history("n2",nx)
  call from_history("o2",ox)
  call from_history("d2",dx)

  call from_param("xsource",xsource)       !position of source in x samples
  call from_param("zsource",zsource)       !position of source in z samples
  call from_param("Ttime",Tt)              !total propagation time
  call from_param("ot",ot,0.0)             !start of seismogram, default=0
  call from_param("fp",fp,25.0)            !Ricker source peak freq.
  call from_param("abc",abc,0)             !# of absorbing border samples
  call from_param("zrec",zrec,1)           !receivers depth
  call from_param("xrec",xrec,(abc+1))     !start of receivers line in x 
  call from_param("dtseis",dtseis,0.002)   !seismogram sampling, default=2ms
  call from_param("lame",lame)             !true for Lame parameters, false for velocities
  call from_param("rec_source",rec_source,.false.)  !this option injects a receiver file into the forward propagation
  call from_param("rec_length",rec_length,(nx - 2*abc)) !length of line of receivers
  call from_param("snap",snap,.false.)     !True if you want to record a movie of the wavefield
  call from_param("randomBound",randomBound,.true.)
  call from_param("write_bkg",write_bkg,.false.) !true if you wanna write the wavefield at the last 2 time steps
  call from_param("nodes",nodes,16)		   !number of nodes for OMP threads

  !In case dt needs to be fixed
  call from_param("dtflag",dtflag,.false.)        !TRUE if you want to read dt and dtratio from file

  if(dtflag) then
    open(unit=20,file="time.txt")
    read(20,*)dt,dtratio
    close(20)
  else
    dt=0
    dtratio=0
  endif

  !OMP
  call omp_set_num_threads(nodes)

  !Division of the model domain in threads
  allocate(b(nodes))
  allocate(e(nodes))
  
  b(1) = 1
  do ix=1,nodes-1
    e(ix) = nx/nodes
    e(ix) = ix*e(ix)
    b(ix+1) = e(ix)
  enddo
  e(nodes) = nx

  !Random Boundaries
  call randomBoundInit()

  allocate(rho(nz,nx))
  if (lame) then
    allocate(lambda(nz,nx))
    allocate(mu(nz,nx))
  else
    allocate(vp(nz,nx))
    allocate(vs(nz,nx))
  endif
  ! Initialize values
  rho    = 0.  
  if (lame) then
    lambda = 0.
    mu     = 0.
  else
    vp  = 0.
    vs  = 0.
  endif
  if (lame) then
    call sep_read(lambda)
    call sep_read(mu,"mu")
  else
    call sep_read(vp)
    call sep_read(vs,"vs")
  endif
  call sep_read(rho,"rho")

  vtemp=0.
  vmax=0.
  if (lame) then
    do ix=1,nx
      do iz=1,nz
        lval=1.0E6*lambda(iz,ix)
        mval=1.0E6*mu(iz,ix)
!         lval=lambda(iz,ix)
!         mval=mu(iz,ix)
        rval=rho(iz,ix)
        vtemp=sqrt((lval+2.0*mval)/rval)
        if (vtemp.gt.vmax) vmax=vtemp
      enddo
    enddo
    if(randomBound) call randomLameBound(lambda,mu,rho)
  else
    vmax = maxval(vp)
    if(randomBound) call randomVelBound(vp,vs,rho)
  endif

  ! dt from the Courant stability condition => C=(vx*dt)/dx+(vz*dt)dz<1
  ! For the simple case of dz=dx and using Virieux(1986) relation:
  ntseis = ceiling(Tt/dtseis) + 1
  if (dt.eq.0.) dt = 0.4*dx/(sqrt(2.0)*vmax) !the 0.4 is arbitrary, but it was diverging otherwise
  if (dtratio.eq.0) dtratio = ceiling(dtseis/dt)
  dt = dtseis/dtratio
  !nt = Tt/dt
  nt = 2*Tt/dt !The 2 appears because in this implementation, each time loop
               !advances only dt/2

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"BEGIN OF SUBROUTINE CALLS"
  write(0,*)"============================================================="
  write(0,*)"Values of the stability conditions:"
  write(0,*)"Max Vp=",vmax
  write(0,*)"dx=dy=",dx
  write(0,*)"dt=",dt
  write(0,*)"lame=",lame
  write(0,*)"dtseis=",dtseis
  write(0,*)"ntseis=",ntseis
  write(0,*)"dtratio=",dtratio
  write(0,*)"============================================================="

  allocate(Vx_rec(rec_length,ntseis))
  allocate(Vz_rec(rec_length,ntseis))
  allocate(sigmaxx_rec(rec_length,ntseis))
  allocate(sigmazz_rec(rec_length,ntseis))
  allocate(sigmaxz_rec(rec_length,ntseis))

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"MEMORY ALLOCATION FOR MODELS AND DATA"
  write(0,*)"============================================================="
  write(0,*)"MODEL"
  if (lame) then
    write(0,*)"rho",size(rho)*4/(1024),"KB"
    write(0,*)"lambda",size(lambda)*4/(1024),"KB"
    write(0,*)"mu",size(mu)*4/(1024),"KB"
  else
    write(0,*)"rho",size(rho)*4/(1024),"KB"
    write(0,*)"Vp",size(vp)*4/(1024),"KB"
    write(0,*)"Vs",size(vs)*4/(1024),"KB"
  endif
  write(0,*)"DATA"
    write(0,*)"Vx",size(Vx_rec)*4/(1024),"KB"
    write(0,*)"Vz",size(Vz_rec)*4/(1024),"KB"
    write(0,*)"sigmaxx",size(sigmaxx_rec)*4/(1024),"KB"
    write(0,*)"sigmazz",size(sigmazz_rec)*4/(1024),"KB"
    write(0,*)"sigmaxz",size(sigmaxz_rec)*4/(1024),"KB"
  write(0,*)"============================================================="

  Vx_rec = 0.
  Vz_rec = 0.
  sigmaxx_rec = 0.
  sigmazz_rec = 0.
  sigmaxz_rec = 0.


  ! Initialize the elastic modeling
  if(lame)      call elastic_init(nz,dz,nx,dx,dt,dtseis,dtratio,xsource,zsource,zrec,xrec,rec_length,fp,nt,ntseis,abc,rho,lambda,mu,b,e,lame,snap,write_bkg)
  if(.not.lame) call elastic_init(nz,dz,nx,dx,dt,dtseis,dtratio,xsource,zsource,zrec,xrec,rec_length,fp,nt,ntseis,abc,rho,vp,vs,b,e,lame,snap,write_bkg)

  if(.not.rec_source) stat = elastic_op(Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  if(rec_source) then
  	allocate(Vx_in(rec_length,ntseis))
  	allocate(Vz_in(rec_length,ntseis))
  	allocate(sigmaxx_in(rec_length,ntseis))
  	allocate(sigmazz_in(rec_length,ntseis))
  	allocate(sigmaxz_in(rec_length,ntseis))
    call sep_read(Vx_in,"Vx_in")
    call sep_read(Vz_in,"Vz_in")
    call sep_read(sigmaxx_in,"sigmaxx_in")
    call sep_read(sigmazz_in,"sigmazz_in")
    call sep_read(sigmaxz_in,"sigmaxz_in")
        
    stat = elastic_op3(Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  endif


  if (stat) then
    write(0,*) "ERROR ON ELASTIC_OP FUNCTION CALL ==> EXITING"
    stop
  else
    write(0,*) "PROPAGATION AND IMAGING CONDITION MODULE ==> OK"
  endif

  deallocate(rho)
  if (lame) deallocate(lambda,mu)
  if(.not.lame)deallocate(vp,vs)

  call seismo(rec_length,dx,ntseis,dtseis,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  write(0,*) "WRITE ==> OK"

  call sep_close()

  end_time = omp_get_wtime()

  ! open(unit=20,file="time.txt")
  ! write(20,*)dt,dtratio
  ! close(20)

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"TOTAL ELAPSED TIME"
  write(0,*)"============================================================="
  write(0,*)"Total=",end_time - start_time, "seconds."
  write(0,*)"============================================================="

end program
