! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 12th, 2015

! RTM code that is supposed to take a forward/adjoint
! input, so that it can be easily tested/put into an
! iterative solver.

module rtm_mod_adj

  use model_mod
  use clib_rtm_ispc
  use clib_lint_ispc
  use clib_image_ispc
  use sep
  use source_mod
  use snap_mod_rtm
  use seis_mod_rtm
  use omp_lib
  use abc_mod

  implicit none

  integer, private :: nt, ntseis, dtratio, nx, nz, xsource, zsource, abc, nf, nm
  real, private    :: dt, dt2, dtseis, dx, dz, fp, aux, aux2, aux3, aux4, fc
  logical, private :: lame, bkg_data, write_bkg, snap
  integer, private :: zrec, xrecbeg, xrecend, xreclen
  integer, private, parameter :: T=1,F=0

  real, pointer, dimension(:,:),private        :: m1,m2,m3
  real, pointer, dimension(:,:),private        :: m1_mig,m2_mig,m3_mig
  real, allocatable, dimension(:,:), private   :: m1_x,m1_z
  real, allocatable, dimension(:,:), private   :: m1_mig_x,m1_mig_z
  real, allocatable, dimension(:,:), private   :: aux_m1_x,aux_m1_z,aux_m2,aux_m3,aux_m2m3
  real, allocatable, dimension(:,:), private   :: aux_m1_mig_x,aux_m1_mig_z,aux_m2_mig,aux_m3_mig,aux_m2m3_mig

  !Density variation for x and z derivatives
  real, allocatable, dimension(:,:),private  :: dm1_x, dm1_z

  !Data
  real, pointer, dimension(:,:),private :: Vx,Vz,sigmaxx,sigmazz,sigmaxz
  !Model
  real, pointer, dimension(:,:),private :: dm1,dm2,dm3
  !Source
  real, allocatable, dimension(:), private   :: source

  !Temporary wavefields
  real, target, allocatable, dimension(:,:)   :: Vx1,Vz1,sigmaxx1,sigmazz1,sigmaxz1
  real, target, allocatable, dimension(:,:)   :: Vx2,Vz2,sigmaxx2,sigmazz2,sigmaxz2

  real, pointer, dimension(:,:), private      :: pVx_tmp,pVz_tmp,psigmaxx_tmp,psigmazz_tmp,psigmaxz_tmp
  real, pointer, dimension(:,:), private      :: pVx_cur,pVz_cur,psigmaxx_cur,psigmazz_cur,psigmaxz_cur
  real, pointer, dimension(:,:), private      :: pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new

  !Background wavefields
  real, target, allocatable, dimension(:,:,:) :: vx0,vz0,sigmaxx0,sigmazz0,sigmaxz0

  real, target, allocatable, dimension(:,:)   :: Vx1b,Vz1b,sigmaxx1b,sigmazz1b,sigmaxz1b
  real, target, allocatable, dimension(:,:)   :: Vx2b,Vz2b,sigmaxx2b,sigmazz2b,sigmaxz2b
  real, target, allocatable, dimension(:,:)   :: Vx3b,Vz3b

  real, pointer, dimension(:,:), private      :: pVx_tmpb,pVz_tmpb,psigmaxx_tmpb,psigmazz_tmpb,psigmaxz_tmpb
  real, pointer, dimension(:,:), private      :: pVx_oldb,pVz_oldb
  real, pointer, dimension(:,:), private      :: pVx_curb,pVz_curb,psigmaxx_curb,psigmazz_curb,psigmaxz_curb
  real, pointer, dimension(:,:), private      :: pVx_newb,pVz_newb,psigmaxx_newb,psigmazz_newb,psigmaxz_newb

  !Imaging condition arrays
  real, target, allocatable, dimension(:,:)   :: s1,s2,s3,s4,s5
  real, pointer, dimension(:,:), private      :: ps1,ps2,ps3,ps4,ps5

  !Nodes thread split
  integer, private                :: nodes
  integer, pointer, dimension(:), private     :: b,e

  contains

!##############################################################################

! Initialization of all variables needed by this subroutine
subroutine rtm_init(nz_in,dz_in,nx_in,dx_in,dt_in,dtseis_in,dtratio_in,xsource_in,zsource_in,zrec_in,xrec_in,rec_length_in,fp_in,nt_in,ntseis_in,abc_in,m1_in,m2_in,m3_in,m1_mig_in,m2_mig_in,m3_mig_in,b_in,e_in,lame_in, bkg_data_in,write_bkg_in,snap_in)

  integer             :: nz_in,nx_in, xsource_in, zsource_in, nt_in, ntseis_in, abc_in,zrec_in,dtratio_in,xrec_in,rec_length_in
  real                :: dz_in,dx_in,dt_in, fp_in, dtseis_in
  logical             :: lame_in, bkg_data_in, write_bkg_in, snap_in
  !Background properties
  real, dimension(:,:),target  :: m1_in,m2_in,m3_in
  real, dimension(:,:),target  :: m1_mig_in,m2_mig_in,m3_mig_in
  integer, dimension(:), target :: b_in, e_in

  nz = nz_in
  dz = dz_in
  dt = dt_in
  dtseis = dtseis_in
  dtratio = dtratio_in
  nx = nx_in
  dx = dx_in
  xsource = xsource_in
  zsource = zsource_in
  fp = fp_in
  nt = nt_in
  ntseis = ntseis_in
  abc = abc_in
  zrec = zrec_in
  xrecbeg = xrec_in
  xreclen = rec_length_in
  lame = lame_in
  bkg_data = bkg_data_in
  write_bkg = write_bkg_in
  snap = snap_in

  xrecend = xrecbeg + xreclen - 1
  
  b => b_in
  e => e_in

  nodes = size(b)
  
  !Input background model
  m1 => m1_in
  m2 => m2_in
  m3 => m3_in
  m1_mig => m1_mig_in
  m2_mig => m2_mig_in
  m3_mig => m3_mig_in

  !Allocate lame parameters
  allocate(m1_x(nz,nx))
  allocate(m1_z(nz,nx))
  allocate(dm1_x(nz,nx))
  allocate(dm1_z(nz,nx))
  allocate(aux_m1_x(nz,nx))
  allocate(aux_m1_z(nz,nx))
  allocate(aux_m2(nz,nx))
  allocate(aux_m3(nz,nx))
  allocate(aux_m2m3(nz,nx))

  allocate(m1_mig_x(nz,nx))
  allocate(m1_mig_z(nz,nx))
  allocate(aux_m1_mig_x(nz,nx))
  allocate(aux_m1_mig_z(nz,nx))
  allocate(aux_m2_mig(nz,nx))
  allocate(aux_m3_mig(nz,nx))
  allocate(aux_m2m3_mig(nz,nx))

  !Allocate source
  allocate(source(nt))
  
  !Allocate Velocities and stresses
  !Temporary wavefields
  allocate(vx0(nz,nx,2))
  allocate(vz0(nz,nx,2))
  allocate(sigmaxx0(nz,nx,2))
  allocate(sigmazz0(nz,nx,2))
  allocate(sigmaxz0(nz,nx,2))

  allocate(Vx1(nz,nx))
  allocate(Vz1(nz,nx))
  allocate(sigmaxx1(nz,nx))
  allocate(sigmazz1(nz,nx))
  allocate(sigmaxz1(nz,nx))
  allocate(Vx2(nz,nx))
  allocate(Vz2(nz,nx))
  allocate(sigmaxx2(nz,nx))
  allocate(sigmazz2(nz,nx))
  allocate(sigmaxz2(nz,nx))
  
  allocate(Vx1b(nz,nx))
  allocate(Vz1b(nz,nx))
  allocate(sigmaxx1b(nz,nx))
  allocate(sigmazz1b(nz,nx))
  allocate(sigmaxz1b(nz,nx))
  allocate(Vx2b(nz,nx))
  allocate(Vz2b(nz,nx))
  allocate(sigmaxx2b(nz,nx))
  allocate(sigmazz2b(nz,nx))
  allocate(sigmaxz2b(nz,nx))
  allocate(Vx3b(nz,nx))
  allocate(Vz3b(nz,nx))
  
  allocate(s1(nz,nx))
  allocate(s2(nz,nx))
  allocate(s3(nz,nx))
  allocate(s4(nz,nx))
  allocate(s5(nz,nx))

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"MEMORY ALLOCATION FOR TEMPORARY ARRAYS"
  write(0,*)"============================================================="
  write(0,*)"INTERPOLATED DENSITY"
  write(0,*)"rho_x",size(m1_x)*4/1024,"KB"
  write(0,*)"rho_z",size(m1_z)*4/1024,"KB"
  write(0,*)"drho_x",size(dm1_x)*4/1024,"KB"
  write(0,*)"drho_z",size(dm1_z)*4/1024,"KB"
  write(0,*)"TEMPORARY ARRAYS"
  write(0,*)"s1",size(s1)*4/(1024*1024),"KB"
  write(0,*)"s2",size(s2)*4/(1024*1024),"KB"
  write(0,*)"s3",size(s3)*4/(1024*1024),"KB"
  write(0,*)"s4",size(s4)*4/(1024*1024),"KB"
  write(0,*)"s5",size(s5)*4/(1024*1024),"KB"
  write(0,*)"Vx0",size(Vx0)*4/1024,"KB"
  write(0,*)"Vz0",size(Vz0)*4/1024,"KB"
  write(0,*)"sigmaxx0",size(sigmaxx0)*4/1024,"KB"
  write(0,*)"sigmazz0",size(sigmazz0)*4/1024,"KB"
  write(0,*)"sigmaxz0",size(sigmaxz0)*4/1024,"KB"
  write(0,*)"Vx1",size(Vx1)*4/1024,"KB"
  write(0,*)"Vz1",size(Vz1)*4/1024,"KB"
  write(0,*)"sigmaxx1",size(sigmaxx1)*4/1024,"KB"
  write(0,*)"sigmazz1",size(sigmazz1)*4/1024,"KB"
  write(0,*)"sigmaxz1",size(sigmaxz1)*4/1024,"KB"
  write(0,*)"Vx2",size(Vx2)*4/1024,"KB"
  write(0,*)"Vz2",size(Vz2)*4/1024,"KB"
  write(0,*)"sigmaxx2",size(sigmaxx2)*4/1024,"KB"
  write(0,*)"sigmazz2",size(sigmazz2)*4/1024,"KB"
  write(0,*)"sigmaxz2",size(sigmaxz2)*4/1024,"KB"
  write(0,*)"Vx1b",size(Vx1)*4/1024,"KB"
  write(0,*)"Vz1b",size(Vz1)*4/1024,"KB"
  write(0,*)"sigmaxx1b",size(sigmaxx1)*4/1024,"KB"
  write(0,*)"sigmazz1b",size(sigmazz1)*4/1024,"KB"
  write(0,*)"sigmaxz1b",size(sigmaxz1)*4/1024,"KB"
  write(0,*)"Vx2b",size(Vx2)*4/1024,"KB"
  write(0,*)"Vz2b",size(Vz2)*4/1024,"KB"
  write(0,*)"sigmaxx2b",size(sigmaxx2)*4/1024,"KB"
  write(0,*)"sigmazz2b",size(sigmazz2)*4/1024,"KB"
  write(0,*)"sigmaxz2b",size(sigmaxz2)*4/1024,"KB"
  write(0,*)"Vx3b",size(Vx3b)*4/1024,"KB"
  write(0,*)"Vz3b",size(Vz3b)*4/1024,"KB"
  write(0,*)"AUXILIARY PROPERTY ARRAYS"
  write(0,*)"aux_m1_x",size(aux_m1_x)*4/1024,"KB"
  write(0,*)"aux_m1_z",size(aux_m1_z)*4/1024,"KB"
  write(0,*)"aux_m2",size(aux_m2)*4/1024,"KB"
  write(0,*)"aux_m3",size(aux_m3)*4/1024,"KB"
  write(0,*)"aux_m2m3",size(aux_m2m3)*4/1024,"KB"
  write(0,*)"============================================================="

  !Initialize values
  !Background properties
  m1_x  = 0.
  m1_z  = 0.
  dm1_x = 0.
  dm1_z = 0.
  m1_mig_x  = 0.
  m1_mig_z  = 0.

  !Source
  source=0.

  !Initialize subroutines
  !Model
  call model_init(nz,nx,dx)
  call model_op(m1,m2,m3,m1_x,m1_z)
  call model_op(m1_mig,m2_mig,m3_mig,m1_mig_x,m1_mig_z)
  call abc_init(nz,nx,dx,abc)

  aux =  dt/dx
  aux3= 1.0/dx
  aux4= 1.0/dt
  aux2= sqrt(float(ntseis)/float(nt))
  dt2 = dt/2.0 !The do loop actually steps at half time steps,
               !so this is needed to sample the seismogram correctly

  !Snapshot
  if(snap) call snap_init(nz,dz,nx,dx,ntseis,dtseis)

  !Source
  call source_init(fp,dt2,dx,nz,nx,nt) !The source is also sampled
                                       !at half time steps

  call source_op2(source)

  aux_m1_x = aux*m1_x
  aux_m1_z = aux*m1_z
  if (lame) then
    aux_m2 = aux*1.0E6*m2
    aux_m3 = aux*1.0E6*m3
    aux_m2m3 = aux*(1.0E6*m2+2.0*1.0E6*m3)
  else
    aux_m2 = aux*m1*(m2*m2-2.0*m3*m3)
    aux_m3 = aux*m1*m3*m3
    aux_m2m3 = aux*m1*m2*m2
  endif

  aux_m1_mig_x = aux*m1_mig_x
  aux_m1_mig_z = aux*m1_mig_z
  if (lame) then
    aux_m2_mig = aux*1.0E6*m2_mig
    aux_m3_mig = aux*1.0E6*m3_mig
    aux_m2m3_mig = aux*(1.0E6*m2_mig+2.0*1.0E6*m3_mig)
  else
    aux_m2_mig = aux*m1_mig*(m2_mig*m2_mig-2.0*m3_mig*m3_mig)
    aux_m3_mig = aux*m1_mig*m3_mig*m3_mig
    aux_m2m3_mig = aux*m1_mig*m2_mig*m2_mig
  endif

  !Background wavefield
  ps1 => s1
  ps2 => s2
  ps3 => s3
  ps4 => s4
  ps5 => s5

end subroutine

!##############################################################################

! Function that is called by Main
function rtm_op(adj,dm1_in,dm2_in,dm3_in,Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in) result(stat)
  logical,intent(in)              :: adj!,add !adj=true is RTM, false is Born
  real,dimension(:,:),target      :: dm1_in,dm2_in,dm3_in
  real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in
  integer                         :: stat

  stat=1
  call rtm_op2(adj,dm1_in,dm2_in,dm3_in,Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in)
  stat=0
end function

!##############################################################################

! Subroutine used by function
subroutine rtm_op2(adj,dm1_in,dm2_in,dm3_in,Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in)

  logical,intent(in)              :: adj!,add
  real,dimension(:,:),target      :: dm1_in,dm2_in,dm3_in
  real,dimension(:,:),target      :: Vx_in,Vz_in,sigmaxx_in,sigmazz_in,sigmaxz_in

  !General variables
  integer                         :: it, ix, iz, itseis

  !Model
  dm1 => dm1_in
  dm2 => dm2_in
  dm3 => dm3_in
  !Data
  Vx => Vx_in
  Vz => Vz_in
  sigmaxx => sigmaxx_in
  sigmazz => sigmazz_in
  sigmaxz => sigmaxz_in

  !###################################################
  ! Check if the backg. propagation needs to be run
  !###################################################
  if (bkg_data.eq..false.) then     

!     !###################################################
!     ! Forward background data d0
!     !###################################################
!     !Time loop for 2nd order time
!     write(0,*)"============================================================="
!     write(0,*)"BEGIN OF TIME MODELING FOR BACKGROUND WAVEFIELD"
!     write(0,*)"============================================================="
!     write(0,*)"Processing:"
!     write(0,*)"0%           25%            50%           75%            100%"
!     write(0,'(a)',advance="no")"["

!     !Zero arrays
!     s1   = 0.
!     s2   = 0.
!     s3   = 0.
!     s4   = 0.
!     s5   = 0.

!     Vx1      = 0.
!     Vz1      = 0.
!     sigmaxx1 = 0.
!     sigmazz1 = 0.
!     sigmaxz1 = 0.
!     Vx2      = 0.
!     Vz2      = 0.
!     sigmaxx2 = 0.
!     sigmazz2 = 0.
!     sigmaxz2 = 0.

!     !Initialize pointers
!     pVx_new => Vx1
!     pVz_new => Vz1
!     psigmaxx_new => sigmaxx1
!     psigmazz_new => sigmazz1
!     psigmaxz_new => sigmaxz1
!     pVx_cur => Vx2
!     pVz_cur => Vz2
!     psigmaxx_cur => sigmaxx2
!     psigmazz_cur => sigmazz2
!     psigmaxz_cur => sigmaxz2

!     b(1)=5
!     e(nodes)=nx-5
  
!     !First time steps
!     !it=1
!     !Boundary condition, all terms are zero

!     !it=2
!     it=2
!     itseis  = it*(dt2/dtseis)

!     call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
!     call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
  
!     !Update pointers
!     pVx_tmp => pVx_cur
!     pVz_tmp => pVz_cur
!     psigmaxx_tmp => psigmaxx_cur
!     psigmazz_tmp => psigmazz_cur
!     psigmaxz_tmp => psigmaxz_cur
!     pVx_cur => pVx_new  
!     pVz_cur => pVz_new 
!     psigmaxx_cur => psigmaxx_new
!     psigmazz_cur => psigmazz_new
!     psigmaxz_cur => psigmaxz_new
!     pVx_new => pVx_tmp
!     pVz_new => pVz_tmp
!     psigmaxx_new => psigmaxx_tmp
!     psigmazz_new => psigmazz_tmp
!     psigmaxz_new => psigmaxz_tmp

!     !it=3
!     it=3
!     itseis = it*(dt2/dtseis)
 
!     !Vx
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
!     enddo
! !$OMP END PARALLEL DO

!     !Vz
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
!     enddo
! !$OMP END PARALLEL DO

!     !Sources
!     call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
!     call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       !sigmaxx
!       call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3)
!       !sigmazz
!       call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2)
!       !sigmaxz
!       call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       !sigmaxx
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2)
!       !sigmazz
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3)
!       !sigmaxz
!       call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3)
!     enddo
! !$OMP END PARALLEL DO

!     !Update pointers
!     pVx_tmp => pVx_cur
!     pVz_tmp => pVz_cur
!     psigmaxx_tmp => psigmaxx_cur
!     psigmazz_tmp => psigmazz_cur
!     psigmaxz_tmp => psigmaxz_cur
!     pVx_cur => pVx_new  
!     pVz_cur => pVz_new 
!     psigmaxx_cur => psigmaxx_new
!     psigmazz_cur => psigmazz_new
!     psigmaxz_cur => psigmaxz_new
!     pVx_new => pVx_tmp
!     pVz_new => pVz_tmp
!     psigmaxx_new => psigmaxx_tmp
!     psigmazz_new => psigmazz_tmp
!     psigmaxz_new => psigmaxz_tmp

!     !it=4 until it=nt
!     do it=4,nt-1
!       itseis  = it*(dt2/dtseis)
    
!       !Vx
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
!       enddo
! !$OMP END PARALLEL DO

!       !Vz
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
!       enddo
! !$OMP END PARALLEL DO
!       call explosive_source(.true.,source(it-1),xsource,zsource,psigmaxx_new)
!       call explosive_source(.true.,source(it-1),xsource,zsource,psigmazz_new)
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!       !sigmaxx
!         call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3) !D
!       !sigmazz
!         call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2) !F
!       !sigmaxz
!         call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3) !H
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!       !sigmaxx
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2) !E
!       !sigmazz
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3) !G
!       !sigmaxz
!         call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3) !J
!       enddo
! !$OMP END PARALLEL DO

!       !Update pointers
!       pVx_tmp => pVx_cur
!       pVz_tmp => pVz_cur
!       psigmaxx_tmp => psigmaxx_cur
!       psigmazz_tmp => psigmazz_cur
!       psigmaxz_tmp => psigmaxz_cur
!       pVx_cur => pVx_new  
!       pVz_cur => pVz_new 
!       psigmaxx_cur => psigmaxx_new
!       psigmazz_cur => psigmazz_new
!       psigmaxz_cur => psigmaxz_new
!       pVx_new => pVx_tmp
!       pVz_new => pVz_tmp
!       psigmaxx_new => psigmaxx_tmp
!       psigmazz_new => psigmazz_tmp
!       psigmaxz_new => psigmaxz_tmp

!       !Write to terminal
!       if (mod(it,(nt/59))==0) then
!         write(0,'(a)',advance="no")"#"
!         close(0)
!       endif

!     enddo

!     write(0,'(a)',advance="yes")"]"

!     if(write_bkg) call seismo2(nz,dz,nx,dx,dt,pvx_new,pvx_cur,pvz_new,pvz_cur,psigmaxx_new,psigmaxx_cur,psigmazz_new,psigmazz_cur,psigmaxz_new,psigmaxz_cur)

  !###################################################
  ! Read the background wavefields from file
  !###################################################
  else
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"READING BACKGROUND WAVEFIELD FROM FILE"
    write(0,*)"============================================================="

    call sep_read(vx0,"vx0")
    call sep_read(vz0,"vz0")
    call sep_read(sigmaxx0,"sigmaxx0")
    call sep_read(sigmazz0,"sigmazz0")
    call sep_read(sigmaxz0,"sigmaxz0")
  endif

  !###################################################
  ! BORN MODELING
  !###################################################
  if(.not.adj) then

!     call shiftL_op2(.false.,.false.,dm1,dm1_x)
!     call shiftU_op2(.false.,.false.,dm1,dm1_z)

!     write(0,*)"============================================================="
!     write(0,*)"BEGIN OF TIME MODELING FOR SCATTERED WAVEFIELD"
!     write(0,*)"============================================================="
!     write(0,*)"Processing:"
!     write(0,*)"0%           25%            50%           75%            100%"
!     write(0,'(a)',advance="no")"["

!     !Zero Arrays
!     Vx1      = 0.
!     Vz1      = 0.
!     sigmaxx1 = 0.
!     sigmazz1 = 0.
!     sigmaxz1 = 0.
!     Vx2      = 0.
!     Vz2      = 0.
!     sigmaxx2 = 0.
!     sigmazz2 = 0.
!     sigmaxz2 = 0.

!     !Zero receiver data
!     Vx = 0.
!     Vz = 0.
!     sigmaxx = 0.
!     sigmazz = 0.
!     sigmaxz = 0.

!     !Initialize pointers
!     pVx_new => Vx1
!     pVz_new => Vz1
!     psigmaxx_new => sigmaxx1
!     psigmazz_new => sigmazz1
!     psigmaxz_new => sigmaxz1
!     pVx_cur => Vx2
!     pVz_cur => Vz2
!     psigmaxx_cur => sigmaxx2
!     psigmazz_cur => sigmazz2
!     psigmaxz_cur => sigmaxz2

!     b(1)=5
!     e(nodes)=nx-5

! ! First time steps
!     !it=1
!     !Boundary condition, all terms are zero
    
!     !it=2
!     it=2
!     itseis  = it*(dt2/dtseis)

! !RIGHT NOW, BORN IS DISABLED
! !REMEMBER TO IMPLEMENTED A BORN SOURCE HERE!!!!

!     if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

! !$OMP SECTIONS
! !$OMP SECTION
!     call lint_serial_c(T,F,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,F,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,F,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP END SECTIONS

!     !Update pointers
!   pVx_tmp => pVx_cur
!   pVz_tmp => pVz_cur
!   psigmaxx_tmp => psigmaxx_cur
!   psigmazz_tmp => psigmazz_cur
!   psigmaxz_tmp => psigmaxz_cur
!   pVx_cur => pVx_new  
!   pVz_cur => pVz_new 
!   psigmaxx_cur => psigmaxx_new
!   psigmazz_cur => psigmazz_new
!   psigmaxz_cur => psigmaxz_new
!   pVx_new => pVx_tmp
!   pVz_new => pVz_tmp
!   psigmaxx_new => psigmaxx_tmp
!   psigmazz_new => psigmazz_tmp
!   psigmaxz_new => psigmaxz_tmp

!     !it=3
!     it=3
!     itseis  = it*(dt2/dtseis)

!     !Vx
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
!     enddo
! !$OMP END PARALLEL DO
!     !Vz
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!       call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!     !sigmaxx
!       call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3)
!     !sigmazz
!       call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2)
!     !sigmaxz
!       call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3)
!     enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!     do ix=1,nodes
!     !sigmaxx
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2)
!     !sigmazz
!       call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3)
!     !sigmaxz
!       call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3)
!     enddo
! !$OMP END PARALLEL DO

!     if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

! !$OMP SECTIONS
! !$OMP SECTION
!     call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!     call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP END SECTIONS

!     !Update pointers
!   pVx_tmp => pVx_cur
!   pVz_tmp => pVz_cur
!   psigmaxx_tmp => psigmaxx_cur
!   psigmazz_tmp => psigmazz_cur
!   psigmaxz_tmp => psigmaxz_cur
!   pVx_cur => pVx_new  
!   pVz_cur => pVz_new 
!   psigmaxx_cur => psigmaxx_new
!   psigmazz_cur => psigmazz_new
!   psigmaxz_cur => psigmaxz_new
!   pVx_new => pVx_tmp
!   pVz_new => pVz_tmp
!   psigmaxx_new => psigmaxx_tmp
!   psigmazz_new => psigmazz_tmp
!   psigmaxz_new => psigmaxz_tmp

!     !it=3 until it=nt
!     do it=4,nt
!       itseis  = it*(dt2/dtseis)

!       !Vx
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Ax_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_cur,pVx_new,aux_m1_x)
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVx_new,aux_m1_x)
!       enddo
! !$OMP END PARALLEL DO
!     !Vz
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Az_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_cur,pVz_new,aux_m1_z)
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!         call Bz_c(F,T,nx,nz,b(ix),e(ix),psigmazz_cur,pVz_new,aux_m1_z)
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!       !sigmaxx
!         call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxx_new,aux_m2m3) !D
!       !sigmazz
!         call Az_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmazz_new,aux_m2) !F
!       !sigmaxz
!         call Bz_c(F,T,nx,nz,b(ix),e(ix),pVx_cur,psigmaxz_new,aux_m3) !H
!       enddo
! !$OMP END PARALLEL DO
! !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
!       do ix=1,nodes
!       !sigmaxx
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxx_new,aux_m2) !E
!       !sigmazz
!         call Bx_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmazz_new,aux_m2m3) !G
!       !sigmaxz
!         call Ax_c(F,T,nx,nz,b(ix),e(ix),pVz_cur,psigmaxz_new,aux_m3) !J
!       enddo
! !$OMP END PARALLEL DO
    
!       if(mod(it,(2*dtratio)).eq.0.and.snap) call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)

! !$OMP SECTIONS
! !$OMP SECTION
!       call lint_serial_c(T,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!       call lint_serial_c(T,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!       call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!       call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP SECTION
!       call lint_serial_c(T,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
! !$OMP END SECTIONS

!       !Update pointers
!     pVx_tmp => pVx_cur
!     pVz_tmp => pVz_cur
!     psigmaxx_tmp => psigmaxx_cur
!     psigmazz_tmp => psigmazz_cur
!     psigmaxz_tmp => psigmaxz_cur
!     pVx_cur => pVx_new  
!     pVz_cur => pVz_new 
!     psigmaxx_cur => psigmaxx_new
!     psigmazz_cur => psigmazz_new
!     psigmaxz_cur => psigmaxz_new
!     pVx_new => pVx_tmp
!     pVz_new => pVz_tmp
!     psigmaxx_new => psigmaxx_tmp
!     psigmazz_new => psigmazz_tmp
!     psigmaxz_new => psigmaxz_tmp

!       !Write to terminal
!       if (mod(it,(nt/59))==0) then
!         write(0,'(a)',advance="no")"#"
!         close(0)
!       endif

!     enddo
!     write(0,'(a)',advance="yes")"]"

  !###################################################
  ! RTM modeling
  !###################################################
  else

    write(0,*)"============================================================="
    write(0,*)"BEGIN OF ADJOINT PROPAGATION OF RECEIVER DATA"
    write(0,*)"============================================================="
    write(0,*)"Processing:"
    write(0,*)"0%           25%            50%           75%            100%"
    write(0,'(a)',advance="no")"["

    !Zero Arrays
    Vx1      = 0.
    Vz1      = 0.
    sigmaxx1 = 0.
    sigmazz1 = 0.
    sigmaxz1 = 0.
    Vx2      = 0.
    Vz2      = 0.
    sigmaxx2 = 0.
    sigmazz2 = 0.
    sigmaxz2 = 0.
    
    Vx1b     = 0.
    Vz1b     = 0.
    sigmaxx1b= 0.
    sigmazz1b= 0.
    sigmaxz1b= 0.
    Vx2b     = 0.
    Vz2b     = 0.
    sigmaxx2b= 0.
    sigmazz2b= 0.
    sigmaxz2b= 0.
    Vx3b     = 0.
    Vz3b     = 0.

    !Initialize pointers
    pVx_new => Vx1
    pVz_new => Vz1
    psigmaxx_new => sigmaxx1
    psigmazz_new => sigmazz1
    psigmaxz_new => sigmaxz1
    pVx_cur => Vx2
    pVz_cur => Vz2
    psigmaxx_cur => sigmaxx2
    psigmazz_cur => sigmazz2
    psigmaxz_cur => sigmaxz2
        
    pVx_newb => Vx1b
    pVz_newb => Vz1b
    psigmaxx_newb => sigmaxx1b
    psigmazz_newb => sigmazz1b
    psigmaxz_newb => sigmaxz1b
    pVx_curb => Vx2b
    pVz_curb => Vz2b
    psigmaxx_curb => sigmaxx2b
    psigmazz_curb => sigmazz2b
    psigmaxz_curb => sigmaxz2b
    
    pVx_oldb => Vx3b
    pVz_oldb => Vz3b

    ! Last time steps
    !it=nt
    !All values are zero

    !it=nt-1
    it=nt-1
    itseis  = it*(dt2/dtseis)

    ! Absorbing boundary condition - START
    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)
    ! Absorbing boundary condition - END

    call lint_serial_c(F,F,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
    call lint_serial_c(F,F,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
    call lint_serial_c(F,F,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
      
    if (mod(it,dtratio).eq.0.and.snap) then
      call snap_op4(pVx_newb,pVz_newb,psigmaxx_newb,psigmazz_newb,psigmaxz_newb)
      call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)
    endif 

    !Update pointers
    pVx_tmp => pVx_cur
    pVz_tmp => pVz_cur
    psigmaxx_tmp => psigmaxx_cur
    psigmazz_tmp => psigmazz_cur
    psigmaxz_tmp => psigmaxz_cur
    pVx_cur => pVx_new  
    pVz_cur => pVz_new 
    psigmaxx_cur => psigmaxx_new
    psigmazz_cur => psigmazz_new
    psigmaxz_cur => psigmaxz_new
    pVx_new => pVx_tmp
    pVz_new => pVz_tmp
    psigmaxx_new => psigmaxx_tmp
    psigmazz_new => psigmazz_tmp
    psigmaxz_new => psigmaxz_tmp

    !it=nt-2
    it=nt-2
    itseis  = it*(dt2/dtseis)

    pVx_newb => vx0(:,:,1)
    pVz_newb => vz0(:,:,1)
    psigmaxx_newb => sigmaxx0(:,:,1)
    psigmazz_newb => sigmazz0(:,:,1)
    psigmaxz_newb => sigmaxz0(:,:,1)
    pVx_curb => vx0(:,:,2)
    pVz_curb => vz0(:,:,2)
    psigmaxx_curb => sigmaxx0(:,:,2)
    psigmazz_curb => sigmazz0(:,:,2)
    psigmaxz_curb => sigmaxz0(:,:,2)

    pVx_oldb => Vx0(:,:,1)!pVx_newb
    pVz_oldb => Vz0(:,:,1)!pVz_newb

    !RECEIVER WAVEFIELD
    b(1)=10
    e(nodes)=nx-9
    !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmaxx_cur,aux_m2m3_mig)
    enddo
!$OMP END PARALLEL DO
    call Azsides_c(T,nx,nz,pVx_new,psigmaxx_cur,aux_m2m3_mig)
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmaxz_cur,aux_m3_mig)
    enddo
!$OMP END PARALLEL DO
    call Bzsides_c(T,nx,nz,pVx_new,psigmaxz_cur,aux_m3_mig)
    b(1)=10
    e(nodes)=nx-9
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmazz_cur,aux_m2_mig)
    enddo
!$OMP END PARALLEL DO
    call Azsides_c(T,nx,nz,pVx_new,psigmazz_cur,aux_m2_mig)
    !Vz
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,F,nx,nz,b(ix),e(ix),pVz_new,psigmaxx_cur,aux_m2_mig)
    enddo
!$OMP END PARALLEL DO
    call Bxsides_c(T,nx,nz,pVz_new,psigmaxx_cur,aux_m2_mig)
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,T,nx,nz,b(ix),e(ix),pVz_new,psigmazz_cur,aux_m2m3_mig)
    enddo
!$OMP END PARALLEL DO
    call Bxsides_c(T,nx,nz,pVz_new,psigmazz_cur,aux_m2m3_mig)
    b(1)=9
    e(nodes)=nx-10
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(T,T,nx,nz,b(ix),e(ix),pVz_new,psigmaxz_cur,aux_m3_mig)
    enddo
!$OMP END PARALLEL DO
    call Axsides_c(T,nx,nz,pVz_new,psigmaxz_cur,aux_m3_mig)
    !sigmaxx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(T,T,nx,nz,b(ix),e(ix),psigmaxx_new,pVx_cur,aux_m1_mig_x)
    enddo
!$OMP END PARALLEL DO
    call Axsides_c(T,nx,nz,psigmaxx_new,pVx_cur,aux_m1_mig_x)
    !sigmazz
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(T,T,nx,nz,b(ix),e(ix),psigmazz_new,pVz_cur,aux_m1_mig_z)
    enddo
!$OMP END PARALLEL DO
    call Bzsides_c(T,nx,nz,psigmazz_new,pVz_cur,aux_m1_mig_z)
    !sigmaxz
    b(1)=10
    e(nodes)=nx-9
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),psigmaxz_new,pVz_cur,aux_m1_mig_z)
    enddo
!$OMP END PARALLEL DO
    call Azsides_c(T,nx,nz,psigmaxz_new,pVz_cur,aux_m1_mig_z)
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,T,nx,nz,b(ix),e(ix),psigmaxz_new,pVx_cur,aux_m1_mig_x)
    enddo
!$OMP END PARALLEL DO
    call Bxsides_c(T,nx,nz,psigmaxz_new,pVx_cur,aux_m1_mig_x)
!$OMP SECTIONS

    ! Absorbing boundary condition - START
    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)
    ! Absorbing boundary condition - END

!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

    !SOURCE WAVEFIELD
    b(1)=5
    e(nodes)=nx-5
    !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Axb_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_curb,pVx_newb,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_curb,pVx_newb,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
  !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Azb_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_curb,pVz_newb,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bzb_c(F,T,nx,nz,b(ix),e(ix),psigmazz_curb,pVz_newb,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
  !sigmaxx
      call Azb_c(F,T,nx,nz,b(ix),e(ix),pVx_curb,psigmaxx_newb,aux_m2m3)
  !sigmazz
      call Azb_c(F,T,nx,nz,b(ix),e(ix),pVx_curb,psigmazz_newb,aux_m2)
  !sigmaxz
      call Bzb_c(F,F,nx,nz,b(ix),e(ix),pVx_curb,psigmaxz_newb,aux_m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmaxx_newb,aux_m2)
    !sigmazz
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmazz_newb,aux_m2m3)
    !sigmaxz
      call Axb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmaxz_newb,aux_m3)
    enddo
!$OMP END PARALLEL DO

    !Imaging condition
    if (mod(it,dtratio).eq.0) then
      b(1)=1
      e(nodes)=nx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call Dt_c(F,F,nx,nz,b(ix),e(ix),pVx_oldb,pVx_newb,ps1,aux4)
        call Dt_c(F,F,nx,nz,b(ix),e(ix),pVz_oldb,pVz_newb,ps2,aux4)
      enddo
!$OMP END PARALLEL DO
      b(1)=5
      e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
      do ix=1,nodes
        call dvxdx_c(F,nx,nz,b(ix),e(ix),pVx_curb,ps3,aux3)
        call dvzdz_c(F,nx,nz,b(ix),e(ix),pVz_curb,ps4,aux3)
        call dvxdz_c(F,nx,nz,b(ix),e(ix),pVx_curb,ps5,aux3)
        call dvzdx_c(T,nx,nz,b(ix),e(ix),pVz_curb,ps5,aux3)
      enddo
!$OMP END PARALLEL DO
      b(1)=1
      e(nodes)=nx
      if (lame) then
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
        do ix=1,nodes
          call image_rho_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps1,ps2,dm1_x,dm1_z,m1_mig_x,m1_mig_z,pVx_cur,pVz_cur)
          call image_lambda_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,dm2,psigmaxx_cur,psigmazz_cur)
          call image_mu_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,ps5,dm3,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
        enddo
!$OMP END PARALLEL DO
      else
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
        do ix=1,nodes
          call image_rho_vel_c(T,F,nx,nz,dt,b(ix),e(ix),ps1,ps2,ps3,ps4,ps5,dm1,dm1_x,dm1_z,m1_mig_x,m1_mig_z,m2_mig,m3_mig,pVx_cur,pVz_cur,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
          call image_vp_vel_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,dm2,m1_mig,m2_mig,psigmaxx_cur,psigmazz_cur)
          call image_vs_vel_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,ps5,dm3,m1_mig,m3_mig,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
        enddo
!$OMP END PARALLEL DO   
      endif
      call shiftL_op2(.true.,.true.,dm1,dm1_x)
      call shiftU_op2(.true.,.true.,dm1,dm1_z)
      if(snap) then
        call snap_op4(pVx_newb,pVz_newb,psigmaxx_newb,psigmazz_newb,psigmaxz_newb)
        call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)
      endif 
    endif
  
      !Update pointers
    pVx_tmp => pVx_cur
    pVz_tmp => pVz_cur
    psigmaxx_tmp => psigmaxx_cur
    psigmazz_tmp => psigmazz_cur
    psigmaxz_tmp => psigmaxz_cur
    pVx_cur => pVx_new  
    pVz_cur => pVz_new 
    psigmaxx_cur => psigmaxx_new
    psigmazz_cur => psigmazz_new
    psigmaxz_cur => psigmaxz_new
    pVx_new => pVx_tmp
    pVz_new => pVz_tmp
    psigmaxx_new => psigmaxx_tmp
    psigmazz_new => psigmazz_tmp
    psigmaxz_new => psigmaxz_tmp

    pVx_tmp => pVx_newb
    pVz_tmp => pVz_newb
    psigmaxx_tmp => psigmaxx_newb
    psigmazz_tmp => psigmazz_newb
    psigmaxz_tmp => psigmaxz_newb
    pVx_newb => pVx_curb
    pVz_newb => pVz_curb
    psigmaxx_newb => psigmaxx_curb
    psigmazz_newb => psigmazz_curb
    psigmaxz_newb => psigmaxz_curb
    pVx_curb => pVx_tmp
    pVz_curb => pVz_tmp 
    psigmaxx_curb => psigmaxx_tmp
    psigmazz_curb => psigmazz_tmp
    psigmaxz_curb => psigmaxz_tmp
  
    !it=nt-3 until it=1
    do it=nt-3,1,-1
      itseis  = it*(dt2/dtseis)

      pVx_oldb = pVx_newb
      pVz_oldb = pVz_newb

      !RECEIVER WAVEFIELD
      b(1)=10
      e(nodes)=nx-9
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmaxx_cur,aux_m2m3_mig)
    enddo
!$OMP END PARALLEL DO
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmaxz_cur,aux_m3_mig)
    enddo
!$OMP END PARALLEL DO
    b(1)=10
    e(nodes)=nx-9
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),pVx_new,psigmazz_cur,aux_m2_mig)
    enddo
!$OMP END PARALLEL DO
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,T,nx,nz,b(ix),e(ix),pVz_new,psigmaxx_cur,aux_m2_mig)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,T,nx,nz,b(ix),e(ix),pVz_new,psigmazz_cur,aux_m2m3_mig)
    enddo
!$OMP END PARALLEL DO
    b(1)=9
    e(nodes)=nx-10
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(T,T,nx,nz,b(ix),e(ix),pVz_new,psigmaxz_cur,aux_m3_mig)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(T,T,nx,nz,b(ix),e(ix),psigmaxx_new,pVx_cur,aux_m1_mig_x)
    enddo
!$OMP END PARALLEL DO
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(T,T,nx,nz,b(ix),e(ix),psigmazz_new,pVz_cur,aux_m1_mig_z)
    enddo
!$OMP END PARALLEL DO
    b(1)=10
    e(nodes)=nx-9
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(T,T,nx,nz,b(ix),e(ix),psigmaxz_new,pVz_cur,aux_m1_mig_z)
    enddo
!$OMP END PARALLEL DO
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(T,T,nx,nz,b(ix),e(ix),psigmaxz_new,pVx_cur,aux_m1_mig_x)
    enddo
!$OMP END PARALLEL DO

!$OMP SECTIONS
!$OMP SECTION
    call Axsides_c(T,nx,nz,psigmaxx_new,pVx_cur,aux_m1_mig_x)
    call Bxsides_c(T,nx,nz,psigmaxz_new,pVx_cur,aux_m1_mig_x)
!$OMP SECTION
    call Bzsides_c(T,nx,nz,psigmazz_new,pVz_cur,aux_m1_mig_z)
    call Azsides_c(T,nx,nz,psigmaxz_new,pVz_cur,aux_m1_mig_z)
!$OMP SECTION
    call Azsides_c(T,nx,nz,pVx_new,psigmaxx_cur,aux_m2m3_mig)
    call Bxsides_c(T,nx,nz,pVz_new,psigmaxx_cur,aux_m2_mig)
!$OMP SECTION
    call Bxsides_c(T,nx,nz,pVz_new,psigmazz_cur,aux_m2m3_mig)       
    call Azsides_c(T,nx,nz,pVx_new,psigmazz_cur,aux_m2_mig)
!$OMP SECTION    
    call Axsides_c(T,nx,nz,pVz_new,psigmaxz_cur,aux_m3_mig)
    call Bzsides_c(T,nx,nz,pVx_new,psigmaxz_cur,aux_m3_mig)
!$OMP END SECTIONS   

    ! Absorbing boundary condition - START
    call K(pVx_new)
    call K(pVz_new)
    call K(psigmaxx_new)
    call K(psigmazz_new)
    call K(psigmaxz_new)
    ! Absorbing boundary condition - END
    
!$OMP SECTIONS
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vx(:,itseis+1:itseis+2),pVx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,Vz(:,itseis+1:itseis+2),pVz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxx(:,itseis+1:itseis+2),psigmaxx_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmazz(:,itseis+1:itseis+2),psigmazz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP SECTION
    call lint_serial_c(F,T,xreclen,dt2,dtseis,sigmaxz(:,itseis+1:itseis+2),psigmaxz_new(zrec,xrecbeg:xrecend),aux2,it)
!$OMP END SECTIONS

  !SOURCE WAVEFIELD
  b(1)=5
    e(nodes)=nx-5
    !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Axb_c(F,T,nx,nz,b(ix),e(ix),psigmaxx_curb,pVx_newb,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_curb,pVx_newb,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
    !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Azb_c(F,T,nx,nz,b(ix),e(ix),psigmaxz_curb,pVz_newb,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bzb_c(F,T,nx,nz,b(ix),e(ix),psigmazz_curb,pVz_newb,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Azb_c(F,T,nx,nz,b(ix),e(ix),pVx_curb,psigmaxx_newb,aux_m2m3) !D
    !sigmazz
      call Azb_c(F,T,nx,nz,b(ix),e(ix),pVx_curb,psigmazz_newb,aux_m2) !F
    !sigmaxz
      call Bzb_c(F,T,nx,nz,b(ix),e(ix),pVx_curb,psigmaxz_newb,aux_m3) !H
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmaxx_newb,aux_m2) !E
    !sigmazz
      call Bxb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmazz_newb,aux_m2m3) !G
    !sigmaxz
      call Axb_c(F,T,nx,nz,b(ix),e(ix),pVz_curb,psigmaxz_newb,aux_m3) !J
    enddo
!$OMP END PARALLEL DO

  !Imaging condition
  if (mod(it,dtratio).eq.0) then
  b(1)=1
  e(nodes)=nx    
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Dt_c(F,F,nx,nz,b(ix),e(ix),pVx_oldb,pVx_newb,ps1,aux4)
      call Dt_c(F,F,nx,nz,b(ix),e(ix),pVz_oldb,pVz_newb,ps2,aux4)
    enddo
!$OMP END PARALLEL DO
    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call dvxdx_c(F,nx,nz,b(ix),e(ix),pVx_curb,ps3,aux3)
      call dvzdz_c(F,nx,nz,b(ix),e(ix),pVz_curb,ps4,aux3)
      call dvxdz_c(F,nx,nz,b(ix),e(ix),pVx_curb,ps5,aux3)
    call dvzdx_c(T,nx,nz,b(ix),e(ix),pVz_curb,ps5,aux3)
    enddo
!$OMP END PARALLEL DO
  b(1)=1
  e(nodes)=nx 
    if(lame) then
!$OMP SECTIONS PRIVATE(IX)
!$OMP SECTION
      do ix=1,nodes
      call image_rho_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps1,ps2,dm1_x,dm1_z,m1_mig_x,m1_mig_z,pVx_cur,pVz_cur)
      enddo
!$OMP SECTION
      do ix=1,nodes
        call image_lambda_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,dm2,psigmaxx_cur,psigmazz_cur)
      enddo
!$OMP SECTION
      do ix=1,nodes
      call image_mu_lame_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,ps5,dm3,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
  enddo
!$OMP END SECTIONS
    else
!$OMP SECTIONS PRIVATE(IX)
!$OMP SECTION
      do ix=1,nodes
      call image_rho_vel_c(T,T,nx,nz,dt,b(ix),e(ix),ps1,ps2,ps3,ps4,ps5,dm1,dm1_x,dm1_z,m1_mig_x,m1_mig_z,m2_mig,m3_mig,pVx_cur,pVz_cur,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
      enddo
!$OMP SECTION
      do ix=1,nodes
      call image_vp_vel_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,dm2,m1_mig,m2_mig,psigmaxx_cur,psigmazz_cur)
      enddo
!$OMP SECTION
      do ix=1,nodes
      call image_vs_vel_c(T,T,nx,nz,dt,b(ix),e(ix),ps3,ps4,ps5,dm3,m1_mig,m3_mig,psigmaxx_cur,psigmazz_cur,psigmaxz_cur)
      enddo
!$OMP END SECTIONS   
    endif
    call shiftL_op2(.false.,.true.,dm1_x,dm1)
    call shiftU_op2(.false.,.true.,dm1_z,dm1)
    if(snap) then
      call snap_op4(pVx_newb,pVz_newb,psigmaxx_newb,psigmazz_newb,psigmaxz_newb)
      call snap_op5(pVx_new,pVz_new,psigmaxx_new,psigmazz_new,psigmaxz_new)
    endif
  endif

    !Update pointers
  pVx_tmp => pVx_cur
  pVz_tmp => pVz_cur
  psigmaxx_tmp => psigmaxx_cur
  psigmazz_tmp => psigmazz_cur
  psigmaxz_tmp => psigmaxz_cur
  pVx_cur => pVx_new  
  pVz_cur => pVz_new 
  psigmaxx_cur => psigmaxx_new
  psigmazz_cur => psigmazz_new
  psigmaxz_cur => psigmaxz_new
  pVx_new => pVx_tmp
  pVz_new => pVz_tmp
  psigmaxx_new => psigmaxx_tmp
  psigmazz_new => psigmazz_tmp
  psigmaxz_new => psigmaxz_tmp

  pVx_tmp => pVx_newb
  pVz_tmp => pVz_newb
  psigmaxx_tmp => psigmaxx_newb
  psigmazz_tmp => psigmazz_newb
  psigmaxz_tmp => psigmaxz_newb
  pVx_newb => pVx_curb
  pVz_newb => pVz_curb
  psigmaxx_newb => psigmaxx_curb
  psigmazz_newb => psigmazz_curb
  psigmaxz_newb => psigmaxz_curb
  pVx_curb => pVx_tmp
  pVz_curb => pVz_tmp 
  psigmaxx_curb => psigmaxx_tmp
  psigmazz_curb => psigmazz_tmp
  psigmaxz_curb => psigmaxz_tmp

    !Write to terminal
    if (mod(it,(nt/59))==0) then
      write(0,'(a)',advance="no")"#"
      close(0)
    endif
  enddo

  write(0,'(a)',advance="yes")"]"

  endif

end subroutine

end module
