! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! This main is to be used with the velocity-stress formulation
! Jan, 12th, 2015

! Main program to execute elastic modeling
program rtm_program

  use sep
  use rtm_mod_adj
  use snap_mod_rtm
  use seis_mod_rtm
  use random_mod

  use omp_lib

  implicit none

  integer                           :: nz,nx,xsource,zsource,nt,abc,zrec,xrec,rec_length,nsnap,ntseis,dtratio
  integer                           :: stat,ix,iz
  real                              :: dz,oz,dx,ox,dt,fp,Tt,ot,dtseis
  real, allocatable, dimension(:,:) :: rho,lambda,mu,vp,vs
  real, allocatable, dimension(:,:) :: rho_mig,lambda_mig,mu_mig,vp_mig,vs_mig
  real, allocatable, dimension(:,:) :: drho,dl,dmu,dvp,dvs
  real, allocatable, dimension(:,:) :: Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec
  real                              :: vmax,vtemp,lval,mval,rval
  logical                           :: snap,adj,lame, bkg_data, write_bkg

  !Variables for omp routines
  real*8                            :: start_time,end_time
  integer                           :: nodes
  integer, allocatable, dimension(:):: b,e

  !OMP settings
  start_time = omp_get_wtime()

  call sep_init()

  ! Read SEP header parameter
  ! Model parameters
  call from_history("n1",nz)
  call from_history("o1",oz)
  call from_history("d1",dz)
  call from_history("n2",nx)
  call from_history("o2",ox)
  call from_history("d2",dx)

  call from_param("nsnap",nsnap,50)        !# of snapshots, default=50
  call from_param("xsource",xsource)       !position of source in x samples
  call from_param("zsource",zsource)       !position of source in z samples
  call from_param("Ttime",Tt)              !total propagation time
  call from_param("ot",ot,0.0)             !start of seismogram, default=0
  call from_param("fp",fp,25.0)            !Ricker source peak freq.
  call from_param("abc",abc,0)             !# of absorbing border samples
  call from_param("zrec",zrec,1)           !receivers depth
  call from_param("xrec",xrec,(abc+1))     !start of receivers line in x 
  call from_param("rec_length",rec_length,(nx - 2*abc)) !length of line of receivers
  call from_param("dtseis",dtseis,0.002)   !seismogram sampling, default=2ms
  call from_param("adj",adj)               !true for RTM, false for Born
  call from_param("lame",lame)             !true for Lame parameters, false for velocities
  call from_param("bkg_data",bkg_data,.false.) !true if you wanna read the data
  call from_param("write_bkg",write_bkg,.false.) !true if you wanna write the data
  call from_param("snap",snap,.false.)     !True if you want to record a movie of the wavefield
  call from_param("randomBound",randomBound,.true.)

  call from_param("nodes",nodes,24) !number of nodes for OMP threads

  !OMP
  call omp_set_num_threads(nodes)

  !TEMPORARY
  !Division of the model domain in threads
  allocate(b(nodes))
  allocate(e(nodes))
  
  b(1) = 1
  do ix=1,nodes-1
    e(ix) = nx/nodes
    e(ix) = ix*e(ix)
    b(ix+1) = e(ix)
  enddo
  e(nodes) = nx

  !Random Boundaries
  call randomBoundInit()

  allocate(rho(nz,nx))
  allocate(drho(nz,nx))
  if (lame) then
    allocate(lambda(nz,nx))
    allocate(mu(nz,nx))
    allocate(dl(nz,nx))
    allocate(dmu(nz,nx))
  else
    allocate(vp(nz,nx))
    allocate(vs(nz,nx))
    allocate(dvp(nz,nx))
    allocate(dvs(nz,nx))
  endif

  !Arrays for properties model of receiver adjoint propagation
  allocate(rho_mig(nz,nx))
  if (lame) then
    allocate(lambda_mig(nz,nx))
    allocate(mu_mig(nz,nx))
  else
    allocate(vp_mig(nz,nx))
    allocate(vs_mig(nz,nx))
  endif


  ! Initialize values
  rho    = 0.  
  rho_mig  = 0.  
  drho   = 0.
  if (lame) then
    lambda   = 0.
    mu       = 0.
    dl       = 0.
    dmu      = 0.
    lambda_mig = 0.
    mu_mig     = 0.
  else
    vp  = 0.
    vs  = 0.
    dvp = 0.
    dvs = 0.
    vp_mig= 0.
    vs_mig= 0.
  endif

  if (lame) then
    call sep_read(lambda)
    call sep_read(mu,"mu")
    call sep_read(lambda_mig,"lambdaM")
    call sep_read(mu_mig,"muM")
  else
    call sep_read(vp)
    call sep_read(vs,"vs")
    call sep_read(vp_mig,"vpM")
    call sep_read(vs_mig,"vsM")
  endif
  call sep_read(rho,"rho")
  call sep_read(rho_mig,"rhoM")

  vtemp=0.
  vmax=0.
  if (lame) then
    do ix=1,nx
      do iz=1,nz
        lval=1.0E6*lambda(iz,ix)
        mval=1.0E6*mu(iz,ix)
        rval=rho(iz,ix)
        vtemp=sqrt((lval+2.0*mval)/rval)
        if (vtemp.gt.vmax) vmax=vtemp
      enddo
    enddo
    if(randomBound) call randomLameBound(lambda,mu,rho)
    if(randomBound) call randomLameBound(lambda_mig,mu_mig,rho_mig)
  else
    vmax = maxval(vp)
    if(randomBound) call randomVelBound(vp,vs,rho)
    if(randomBound) call randomVelBound(vp_mig,vs_mig,rho_mig)
  endif

  ! dt from the Courant stability condition => C=(vx*dt)/dx+(vz*dt)dz<1
  ! For the simple case of dz=dx and using Virieux(1986) relation:
  ntseis = ceiling(Tt/dtseis) + 1
  dt = 0.4*dx/(sqrt(2.0)*vmax) !the 0.4 is arbitrary, but it was diverging otherwise
  dtratio = ceiling(dtseis/dt)
  dt = dtseis/dtratio
  !nt = Tt/dt
  nt = 2*Tt/dt !The 2 appears because in this implementation, each time loop
               !advances only dt/2

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"BEGIN OF SUBROUTINE CALLS"
  write(0,*)"============================================================="
  write(0,*)"Values of the stability conditions:"
  write(0,*)"Max Vp=",vmax
  write(0,*)"dx=dy=",dx
  write(0,*)"dt=",dt
  write(0,*)"lame=",lame
  write(0,*)"dtseis=",dtseis
  write(0,*)"ntseis=",ntseis
  write(0,*)"dtratio=",dtratio
  write(0,*)"============================================================="

  write(0,*) 'Receiver line : ', rec_length
  allocate(Vx_rec(rec_length,ntseis))
  allocate(Vz_rec(rec_length,ntseis))
  allocate(sigmaxx_rec(rec_length,ntseis))
  allocate(sigmazz_rec(rec_length,ntseis))
  allocate(sigmaxz_rec(rec_length,ntseis))

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"MEMORY ALLOCATION FOR MODELS AND DATA"
  write(0,*)"============================================================="
  write(0,*)"BACKGROUND"
  if (lame) then
    write(0,*)"rho",size(rho)*4/(1024),"KB"
    write(0,*)"lambda",size(lambda)*4/(1024),"KB"
    write(0,*)"mu",size(mu)*4/(1024),"KB"
    write(0,*)"MIGRATION MODEL"
    write(0,*)"rhoM",size(rho_mig)*4/(1024),"KB"
    write(0,*)"lambdaM",size(lambda_mig)*4/(1024),"KB"
    write(0,*)"muM",size(mu_mig)*4/(1024),"KB"
  else
    write(0,*)"rho",size(rho)*4/(1024),"KB"
    write(0,*)"Vp",size(vp)*4/(1024),"KB"
    write(0,*)"Vs",size(vs)*4/(1024),"KB"
    write(0,*)"MIGRATION MODEL"
    write(0,*)"rhoM",size(rho_mig)*4/(1024),"KB"
    write(0,*)"VpM",size(vp_mig)*4/(1024),"KB"
    write(0,*)"VsM",size(vs_mig)*4/(1024),"KB"
  endif
  write(0,*)"IMAGE"
  if (lame) then
    write(0,*)"drho",size(drho)*4/(1024),"KB"
    write(0,*)"dlambda",size(dl)*4/(1024),"KB"
    write(0,*)"dmu",size(dmu)*4/(1024),"KB"
  else
    write(0,*)"drho",size(drho)*4/(1024),"KB"
    write(0,*)"dVp",size(dvp)*4/(1024),"KB"
    write(0,*)"dVs",size(dvs)*4/(1024),"KB"
  endif
  write(0,*)"DATA"
    write(0,*)"Vx",size(Vx_rec)*4/(1024),"KB"
    write(0,*)"Vz",size(Vz_rec)*4/(1024),"KB"
    write(0,*)"sigmaxx",size(sigmaxx_rec)*4/(1024),"KB"
    write(0,*)"sigmazz",size(sigmazz_rec)*4/(1024),"KB"
    write(0,*)"sigmaxz",size(sigmaxz_rec)*4/(1024),"KB"
  write(0,*)"============================================================="

  Vx_rec = 0.
  Vz_rec = 0.
  sigmaxx_rec = 0.
  sigmazz_rec = 0.
  sigmaxz_rec = 0.

  if (adj) then
    call sep_read(Vx_rec,"Vx_rec")
    call sep_read(Vz_rec,"Vz_rec")
    call sep_read(sigmaxx_rec,"sigmaxx_rec")
    call sep_read(sigmazz_rec,"sigmazz_rec")
    call sep_read(sigmaxz_rec,"sigmaxz_rec")
    write(0,*)"READ SEISMOGRAM => OK"
  else
    if (lame) then
      call sep_read(dl,"dlambda")
      call sep_read(dmu,"dmu")
    else
      call sep_read(dvp,"dvp")
      call sep_read(dvs,"dvs")
    endif
    call sep_read(drho,"drho")
  endif

  ! Initialize the elastic modeling
  !call rtm_init(nz,dz,nx,dx,dt,xsource,zsource,zrec,fp,nt,abc,free_surface,lambda,mu,rho)
  if (lame) call rtm_init(nz,dz,nx,dx,dt,dtseis,dtratio,xsource,zsource,zrec,xrec,rec_length,fp,nt,ntseis,abc,rho,lambda,mu,rho_mig,lambda_mig,mu_mig,b,e,lame, bkg_data,write_bkg,snap)
  if (.not.lame) call rtm_init(nz,dz,nx,dx,dt,dtseis,dtratio,xsource,zsource,zrec,xrec,rec_length,fp,nt,ntseis,abc,rho,vp,vs,rho_mig,vp_mig,vs_mig,b,e,lame, bkg_data,write_bkg,snap)

  write(0,*) "INITIALIZATION ==> OK"

  !stat = rtm_op(adj,dl,dmu,drho,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  if (lame) stat = rtm_op(adj,drho,dl,dmu,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  if (.not.lame) stat = rtm_op(adj,drho,dvp,dvs,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)

  if (stat) then
    write(0,*) "ERROR ON RTM_OP FUNCTION CALL ==> EXITING"
    stop
  else
    write(0,*) "PROPAGATION AND IMAGING CONDITION MODULE ==> OK"
  endif

  if (.not.adj) call seismo(rec_length,dx,ntseis,dtseis,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  !if (.not.adj) call seismo(nx,dx,nt,dt,Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  if (adj) then
    if (lame) then
      call img(nx,dx,nz,dz,drho,dl,dmu,lame)
    else
      call img(nx,dx,nz,dz,drho,dvp,dvs,lame)
    endif
  endif
  write(0,*) "WRITE ==> OK"

  call sep_close()

!   deallocate(rho,drho)
!   if (lame) deallocate(lambda,mu,dl,dmu)
!   if(.not.lame)deallocate(vp,vs,dvp,dvs)

  end_time = omp_get_wtime()

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"TOTAL ELAPSED TIME"
  write(0,*)"============================================================="
  write(0,*)"Total=",end_time - start_time, "seconds."
  write(0,*)"============================================================="

end program
